// SPDX-FileCopyrightText: 2022 Daniel Vrátil <dvratil@kde.org>
//
// SPDX-License-Identifier: MIT

#pragma once

namespace QCoro {

template<typename T> class Task;
template<typename T> class Generator;
template<typename T> class GeneratorIterator;
template<typename T> class AsyncGenerator;
template<typename T> class AsyncGeneratorIterator;

} // namespace QCoro
